/*
 * Copyright (c) 2003-2006 Erez Zadok
 * Copyright (c) 2003-2006 Stony Brook University
 * Copyright (c) 2003-2006 The Research Foundation of State University of New York
 * Copyright (c) 2004-2006 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2006 Gopala Surya
 * Copyright (c) 2004-2005 Tim Wong
 */

#include "context_hash.h"
#include <string.h>
#include <assert.h>

#define DEFAULT_TABLESIZE 1024

typedef struct context_hash_entry {
	struct context_hash_entry* next;
	void* context;
	int id;
	int op;
} context_hash_entry;

static context_hash_entry* hash_table[DEFAULT_TABLESIZE];

static inline int __context_hash_function(int id)
{
	return id % DEFAULT_TABLESIZE;
}

static int __is_ignore(int op)
{
	if (op == OP_READ || op == OP_READPAGE ||
	    op == OP_WRITE_INODE || op == OP_PUT_INODE ||
	    op == MSG_STOP || op == MSG_START || op == MSG_PNAME ||
	    op == OP_DELETE_INODE || op == OP_CLEAR_INODE ||
	    op == OP_D_IPUT)
		return 1;
	return 0;
}

int context_hash_add(int id, int op, void* context)
{
	int ret = 0;
	int index = __context_hash_function(id);
	context_hash_entry* entry;
	context_hash_entry* ptr = hash_table[index];

	print_entry_location();

	if (__is_ignore(op))
		goto exit;

	if (context == NULL) {
		ret = -1;
		goto exit;
	}

	entry = malloc(sizeof(context_hash_entry));
	if(entry == NULL) {
		ret = -1;
		print_error_nomem();
		goto exit;
	}

	entry->next = NULL;
	entry->id = id;
	entry->op = op;
	entry->context = context;

	if (ptr == NULL) {
		hash_table[index] = entry;
		ret = 0;
		goto exit;
	}

	while (ptr->next != NULL) {
		assert(ptr->id != id || ptr->op != op);
		ptr = ptr->next;
	}
	ptr->next = entry;

exit:
	print_exit_location();
	return ret;
}

void* context_hash_remove(int id, int op)
{
	void* ret = NULL;
	context_hash_entry* prev = NULL;
	context_hash_entry* ptr;
	int index;

	print_entry_location();
	if (__is_ignore(op))
                goto out_ok;

	index = __context_hash_function(id);
	ptr = hash_table[index];
	while (ptr != NULL) {
		if (ptr->id == id && ptr->op == op) {
			if (prev != NULL)
				prev->next = ptr->next;
			else {
				if (ptr->next == NULL)
					hash_table[index] = NULL;
				else
					hash_table[index] = ptr->next;
			}
			ret = ptr->context;
			free(ptr);
			goto out_ok;
		}
		prev = ptr;
		ptr = ptr->next;
	}
out:
	fprintf(stderr, "Warning: Failed attempt to free: ID: %d OP: %d\n", id, op);
out_ok:
	print_exit_location();
	return ret;
}

int context_hash_init()
{
	int ret = 0;
	print_entry_location();

	memset(hash_table, 0, sizeof(context_hash_entry*) * DEFAULT_TABLESIZE);

	print_exit_location();
	return ret;
}

inline int context_hash_cleanup()
{
	context_hash_entry* ptr = NULL;
	context_hash_entry* prev = NULL;
	int index = 0;

	print_entry_location();

	for (index = 0; index < DEFAULT_TABLESIZE; index++) {
		ptr = hash_table[index];
		while (ptr != NULL) {
			fprintf(stderr, "Warning: Unfreed SeqNum hash entry: ID: %d OP: %d\n", ptr->id, ptr->op);
			prev = ptr;
			ptr = ptr->next;
			free(prev);
		}
		hash_table[index] = NULL;
	}

	print_exit_location();
	return 0;
}

