/*
 * Copyright (c) 2003-2006 Erez Zadok
 * Copyright (c) 2003-2006 Stony Brook University
 * Copyright (c) 2003-2006 The Research Foundation of State University of New York
 * Copyright (c) 2004-2006 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2006 Gopala Surya
 */

#include "idriver.h"
#include <openssl/hmac.h>
#include "common.h"
#include "od_cksum.h"
#include <string.h>

static int buf_size;
static char *buf;
static HMAC_CTX hmac_ctx;
static const EVP_MD *md;
static unsigned char hmac_value[EVP_MAX_MD_SIZE];
static unsigned int hmac_len;
static int read;
static int remaining;
static block_index_t block_index = 0;


static int cdriver_init(struct parser *p, void *params, struct idriver *next)
{
	struct cksum_params *cparams = (struct cksum_params *)params;
	if (!next) {
		snprintf(p->err, ERR_SIZE, "cksum driver requires a next driver");
		return ERR_INVAL;
	}
	cksum_idriver->next_driver = next;
	if (!cparams || cparams->keylen > HMAC_MAX_KEY_LENGTH) {
		snprintf(p->err, ERR_SIZE, "Invalid params for cksum driver");
		return ERR_INVAL;
	}
	buf_size = cparams->blk_size;
	if (!buf_size)
		buf_size = PAGE_SIZE;
	if ((buf = malloc(buf_size)) == 0) {
		snprintf(p->err, ERR_SIZE, "Could not allocate buffer for cksum driver");
		return ERR_NOMEM;
	}
	md = EVP_md5();
	HMAC_CTX_init(&hmac_ctx);
	HMAC_Init(&hmac_ctx, cparams->key, cparams->keylen, md);
	return 0;

}

static int cdriver_read(struct parser *p, char *buffer, int len)
{
	int total = len;
	block_index_t current;
	char *ptr = buffer;
	while (remaining < total) {
		if (remaining) {
			memcpy(ptr, buf + read - remaining, remaining);
			total -= remaining;
			ptr += remaining;
		}
		read = cksum_idriver->next_driver->ops->read(p, buf, buf_size);
		if (read < 0) {
			snprintf(p->err, ERR_SIZE, "Error while reading file: %d", errno);
			return read;
		}
		remaining = read;
		if (remaining == 0)
			break;
		if (read < EVP_MD_size(md) + sizeof(block_index)) {
			snprintf(p->err, ERR_SIZE, "Checksum not found in file");
			return -ERR_INCOMPLETE;
		}
		remaining -= EVP_MD_size(md);
		HMAC_Init_ex(&hmac_ctx, NULL, 0, md, NULL);
		HMAC_Update(&hmac_ctx,(const unsigned char *) buf + EVP_MD_size(md), remaining);
		HMAC_Final(&hmac_ctx, hmac_value, &hmac_len);
		if (memcmp(buf, hmac_value, hmac_len) != 0) {
			snprintf(p->err, ERR_SIZE, "Checksum failed for block: %u", block_index);
			return -ERR_CKSUM_FAILED;
		}
		memcpy(&current, buf + EVP_MD_size(md), sizeof(current));
		if (current != block_index) {
			snprintf(p->err, ERR_SIZE, "Missing block index, looking for %u, found %u\n", block_index, current);
			return -ERR_BLOCK_MISMATCH;
		}
		remaining -= sizeof(block_index);
		block_index++;
	}
	if (total && remaining > 0) {
		memcpy(ptr, buf + read - remaining, total);
		remaining -= total;
		total = 0;
	}

	return (len - total);

}

static void cdriver_release()
{
	HMAC_CTX_cleanup(&hmac_ctx);
	free(buf);
}

struct input_ops cdriver_ops  = {
	init: cdriver_init,
	read: cdriver_read,
	release: cdriver_release
};

struct idriver cdriver = {
	ops : &cdriver_ops
};

struct idriver *cksum_idriver = &cdriver;
