/*
 * Copyright (c) 2003-2006 Erez Zadok
 * Copyright (c) 2003-2006 Stony Brook University
 * Copyright (c) 2003-2006 The Research Foundation of State University of New York
 * Copyright (c) 2004-2006 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2006 Gopala Surya
 */

#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netinet/in.h>

#include "tracereader.h"
#include "context_hash.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include "stream_header.h"
#include "idriver.h"
#include "hash.h"
#include <assert.h>

#define ARG_DESC_HASH_SIZE 100

static int get_len_from_msg_type(hash_t* hash,field_t type);
static int fill_msg_type_hash(hash_t* hash);

static hash_t arg_desc_hash;

int process_trace_message(struct parser *p, char *data, int size)
{
	char *ptr = data;
	int err = ERR_OK;
	void *buf = NULL;
	field_t arg_type;
	int len = -1;
	int new_len;
	field_len_t length;
	while (err == ERR_OK && ptr < data + size)
	{
		switch(p->state) {
			case IN_MSG:
				buf = (void *)&arg_type;
				len = sizeof(arg_type);
				break;
			case IN_ARG_TLV_GETLEN:
				buf = (void *)&length;
				len = sizeof(length);
				break;
			case IN_ARG_TLV:
			case IN_ARG_TV:
				/* nothing to do here, len is already set */
				break;
			default:
				on_error(ERR_INVAL, p->msg_num, "Unknown state");
				break;

		}

		if (ptr + len > data + size) {
			err = ERR_INVALID_LENGTH;
			break;
		}
		new_len = len;
		switch (p->state) {
			case IN_MSG:
				memcpy(buf, ptr, len);
				if (arg_type & FIELD_LENGTH_MASK) {
					arg_type &= ~FIELD_LENGTH_MASK;
					p->state = IN_ARG_TLV_GETLEN;
				} else {
					if ((new_len = get_len_from_msg_type(&arg_desc_hash,arg_type)) < 0) {
						printf("Offending arg_type = %d\n", arg_type);
						err = ERR_INVALID_TYPE;
					}
					p->state = IN_ARG_TV;
				}
				p->current_arg_type = arg_type;
				break;
			case IN_ARG_TLV_GETLEN:
				memcpy(buf, ptr, len);
				new_len = length;
				if (len < 0) {
					err = ERR_INVALID_LENGTH;
				} else if (len > 0) {
					p->state = IN_ARG_TLV;
				} else {
					p->state = IN_MSG;
				}
				break;
			case IN_ARG_TV:
				err = on_data_field(p->current_msg_type, p->pre_or_post, 0, p->current_arg_type, len, ptr, p->context);
				p->state = IN_MSG;
				break;
			case IN_ARG_TLV:
				err = on_data_field(p->current_msg_type, p->pre_or_post, 1, p->current_arg_type, len, ptr, p->context);
				p->state = IN_MSG;
				break;
			default:
				on_error(ERR_INVAL, p->msg_num, "Unknown state");
				break;
		}
		ptr += len;
		len = new_len;

	}
	if (err == ERR_OK)
		err = on_end_trace_message(p->current_msg_type, p->pre_or_post, p->context);
	p->context = NULL;
	p->msg_num++;
	return err;
}

static void release_all_drivers(struct parser *p)
{
	struct idriver *tmp = p->idriver;
	while (tmp) {
		tmp->ops->release();
		tmp = tmp->next_driver;
	}
}

static int initialize_reader(struct parse_params *params, struct parser *p)
{
	int i;
	int err = ERR_OK;
	struct idriver *tmp = NULL;
	memset(p, 0, sizeof(*p));

	context_hash_init();
	hash_init(&arg_desc_hash,ARG_DESC_HASH_SIZE);
	fill_msg_type_hash(&arg_desc_hash);

	if (params->n_ops < 1 || params->n_ops > MAX_OPS) {
		snprintf(p->err, ERR_SIZE, "Invalid number of ops specified\n");
		return ERR_INVAL;
	}
	for (i = params->n_ops - 1; i >= 0; i--) {

		switch(params->ops[i]) {
			case OP_CKSUM_MD5:
				tmp = cksum_idriver;
				err = cksum_idriver->ops->init(p, (void
							*)&(params->cparams), p->idriver);
				break;
			case OP_FILE:
				tmp = file_idriver;
				err = file_idriver->ops->init(p, (void *)params->file_name, p->idriver);
				break;
			case OP_UNCOMPRESS:
				tmp = uncompress_idriver;
				err = uncompress_idriver->ops->init(p, NULL, p->idriver);
				break;
			case OP_DECRYPT:
				tmp = decrypt_idriver;
				err = decrypt_idriver->ops->init(p, (void *)&(params->dparams), p->idriver);
				break;
		}
		if (err != ERR_OK) {
			release_all_drivers(p);
			break;
		}
		p->idriver = tmp;
	}
	return err;


}


/* main library function, calls all the callbacks */
int parse(struct parse_params *params) {
	int err = ERR_OK;
	struct parser p;
	struct stream_header header;
	int nread;
	void *buf = NULL;
	int len = 0;
	int done = 0;
     	msg_len_t msg_len;
	op_t msg_type;
	int msg_seqnum;
	params->bytes_read = 0;
	memset(&p, 0, sizeof(p));
	if ((err = initialize_reader(params, &p)) != ERR_OK)
	    goto out;
	nread = p.idriver->ops->read(&p, (char *)&header, sizeof(header));
	if (nread > 0 && nread < sizeof(header)) {
	    strncpy(p.err, "Header not found", ERR_SIZE);
	    err = ERR_INCOMPLETE;
	    goto out;
	} else if (nread < 0) {
	    err = -nread;
	    goto out;
	}
	params->bytes_read += nread;
	if (ntohl(header.magic) != TRACEFS_STREAM_MAGIC) {
	    strncpy(p.err, "Invalid magic number, not a trace file", ERR_SIZE);
	    err = ERR_INVAL;
	    goto out;
	}
	if (ntohl(header.version) != VERSION_CODE) {
	    version_t file_ver = ntohl(header.version);
	    if (MAJOR_VERSION(file_ver) != MAJOR_VERSION(VERSION_CODE)) {
		strncpy(p.err, "Version mismatch. Cannot continue.", ERR_SIZE);
		err = ERR_VERSION_MISMATCH;
		goto out;
	    }
	    if (MINOR_VERSION(file_ver) > MINOR_VERSION(VERSION_CODE)) {
		char buf[256];
		snprintf(buf, sizeof(buf), "File version: %lu.%lu.%lu, parser version %s",
			 MAJOR_VERSION(file_ver), MINOR_VERSION(file_ver), REVISION(file_ver),
			 VERSION_STR);
		on_warning(WARN_VERSION, buf);
	    }
	}
	if ((err = on_header(&header)) != ERR_OK)
		goto out;


	/* read the header from the file */
	p.context = NULL;
	p.state = BEGIN;
	while (!done) {
		switch (p.state) {
			case BEGIN:
				buf = (void *)&msg_type;
				len = sizeof(msg_type);
				break;
			case IN_MSG_LEN:
				buf = (void *)&msg_len;
				len = sizeof(msg_len);
				break;
		        case IN_MSG_SEQ_NUM:
			        buf = (void*)&msg_seqnum;
				len = sizeof(msg_seqnum);
				break;
			case IN_MSG_READ:
			        buf = malloc(msg_len);
				len = msg_len;
				if (!buf)
					err = ERR_NOMEM;
				break;
			default:
				on_error(ERR_INVAL, p.msg_num, "Unknown state");
				break;

		}
		if (err != ERR_OK)
			break;
		nread = p.idriver->ops->read(&p, buf, len);
		if (nread > 0 && nread < len) {
			err = ERR_INCOMPLETE;
			break;
		} else if (nread <= 0) {
			err = -nread;
			break;
		}

		params->bytes_read += nread;
		switch(p.state) {
			case BEGIN:
				p.current_msg_type = GET_OP(msg_type);
				p.pre_or_post = IS_POST_OP(msg_type);
				p.state = IN_MSG_LEN;
				break;
			case IN_MSG_LEN:
				p.state = IN_MSG_SEQ_NUM;
				break;
		        case IN_MSG_SEQ_NUM:
			        p.sequence_number = msg_seqnum;
				//fprintf(stderr, "SEQNUM: %d, POST: %d, TYPE: %d\n", p.sequence_number, p.pre_or_post, p.current_msg_type);
				if(p.pre_or_post) /* post */
					p.context = context_hash_remove(p.sequence_number, p.current_msg_type);
				p.context = on_start_trace_message(p.current_msg_type, p.pre_or_post, p.context);
				if(!p.pre_or_post) /* pre */
					context_hash_add(p.sequence_number, p.current_msg_type, p.context);
				p.state = IN_MSG_READ;
				break;
			case IN_MSG_READ:
				p.state = IN_MSG;
				err = process_trace_message(&p, buf, msg_len);
				free(buf);
				p.state = BEGIN;
				if (p.current_msg_type == MSG_STOP)
					done = 1;
				break;
			default:
				on_error(ERR_INVAL, p.msg_num, "Unknown state");
				break;
		}
		if (err != ERR_OK || nread == 0)
			break;
	}
	if (p.state != BEGIN) {
		err = ERR_INCOMPLETE;
	}

out:
	release_all_drivers(&p);
	context_hash_cleanup();

	if (err != ERR_OK) {
		on_error(err, p.msg_num, (p.err[0] == '\0') ? "Error!" : p.err);
	}
	return err;
}

static int get_len_from_msg_type(hash_t* hash,field_t arg_type)
{
	int* ret = hash_find(hash,&arg_type,sizeof(arg_type));
	if(ret == NULL)
		return -1;
	return *ret;
}

static int fill_msg_type_hash(hash_t* hash)
{
	int i;

	assert(hash != NULL);
	for(i = 0; arg_desc[i].str_type != NULL; i++)
		assert(!hash_add(hash,
				 &arg_desc[i].arg_type,
				 sizeof(arg_desc[i].arg_type),
				 &arg_desc[i].size));
	return 0;
}
